# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

from struct import unpack, pack
from pyDiameter.pyDiaAVPBase import DiaAVP

from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_INT32_BUFF_LEN
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_UINT32_BUFF_LEN
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_INT64_BUFF_LEN
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_UINT64_BUFF_LEN
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_FLOAT32_BUFF_LEN
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_VALUE_FLOAT64_BUFF_LEN
from pyDiameter.pyDiaAVPConst import AVP_LENGTH_BUFF_LEN
from pyDiameter.pyDiaConst import ZERO_LEN_BYTES
from pyDiameter.pyDiaConst import DEFAULT_BYTES_ORDER_2
from pyDiameter.pyDiaAVPFactory import DiaAVPFactory
from pyDiameter.pyDiaAVPPath import DiaAVPPath

# An unknown type of Diameter AVP.																						#
class DiaAVPUnknowType(DiaAVP):	
	def _decodeAVPValue(self, valueBytes):
		return None
		
	def _encodeAVPValue(self, avpValue):
		return ZERO_LEN_BYTES
		
	def _validAVPValue(self, value):
		return True
		
	def __len__(self):
		return 0
	
	def _initAVPValue(self):
		pass
		
	def __getitem__(self, index):
		pass
		
# end of DiaAVPUnknowType.
		
class DiaAVPStr(DiaAVP):	
	# this function will decode bytes value to value.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		# in string type AVP, valueBytes will be return directly.
		return valueBytes
		
	# Override the encode function.																						#
	def _encodeAVPValue(self, avpValue):
		assert type(avpValue) is bytes, "Make sure avpValue is Python Bytes Type. Now it is " + str(type(avpValue)) 
		return avpValue
		
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if not type(value) is bytes: raise ValueError("String AVP should have a bytes-like value. e.g. b'hello, world.'")
		return True
		
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(ZERO_LEN_BYTES)
		
	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + len(self.getAVPValue())
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPStr')
		
# end of DiaAVPStr.
		
class DiaAVPUInt32(DiaAVP):		
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_UINT32_BUFF_LEN, "For UINT32 AVP, buffer should have 4 bytes, now it is " + str(len(valueBytes))
		return self.decodeUIntValue(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			return self.encodeUIntValue(avpValue, AVP_VALUE_UINT32_BUFF_LEN)
		except OverflowError as ofle:
			raise OverflowError("UINT32 AVP should be b/w 0 - 4294967295, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is int) or (value < 0) : raise ValueError("UINT32 AVP should have a positive integer value. e.g. 10")
		return True
		
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0)

	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_UINT32_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPUInt32.')
			
# end of DiaAVPUInt32.

class DiaAVPInt32(DiaAVP):		
	def __decodeInt32Value(self, valueBytes):
		return unpack(DEFAULT_BYTES_ORDER_2+'i', valueBytes)[0]
	
	def __encodeInt32Value(self, value):
		try:
			return pack(DEFAULT_BYTES_ORDER_2+'i', value)
		except:
			raise OverflowError
		
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_INT32_BUFF_LEN, "For INT32 AVP, buffer should have 4 bytes, now it is " + str(len(valueBytes))
		return self.__decodeInt32Value(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			return self.__encodeInt32Value(avpValue)
		except OverflowError as ofle:
			raise OverflowError("INT32 AVP should be b/w -2147483648 - 2147483647, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is int): raise ValueError("INT32 AVP should have a integer value. e.g. 10, -10")
		return True
		
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0)
		
	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_INT32_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPInt32.')

# end of DiaAVPInt32

class DiaAVPUInt64(DiaAVP):
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_UINT64_BUFF_LEN, "For UINT64 AVP, buffer should have 8 bytes, now it is " + str(len(valueBytes))
		return self.decodeUIntValue(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			return self.encodeUIntValue(avpValue, AVP_VALUE_UINT64_BUFF_LEN)
		except OverflowError as ofle:
			raise OverflowError("UINT64 AVP should be b/w 0 - 18446744073709551615, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is int) or (value < 0) : raise ValueError("UINT64 AVP should have a positive integer value. e.g. 10")
		return True
			
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0)

	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_UINT64_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPUInt64.')
		
# end of DiaAVPUInt64.

class DiaAVPInt64(DiaAVP):
	def __decodeInt64Value(self, valueBytes):
		return unpack(DEFAULT_BYTES_ORDER_2+'q', valueBytes)[0]
	
	def __encodeInt64Value(self, value):
		try:
			return pack(DEFAULT_BYTES_ORDER_2+'q', value)
		except:
			raise OverflowError
		
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_INT64_BUFF_LEN, "For INT64 AVP, buffer should have 8 bytes, now it is " + str(len(valueBytes))
		return self.__decodeInt64Value(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			return self.__encodeInt64Value(avpValue)
		except OverflowError as ofle:
			raise OverflowError("INT64 AVP should be b/w -9223372036854775808 - 9223372036854775807, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is int): raise ValueError("INT32 AVP should have a integer value. e.g. 10, -10")
		return True
			
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0)
		
	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_INT64_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPInt64.')
		
# end of DiaAVPInt64
	
class DiaAVPFloat32(DiaAVP):
	def __decodeFloat32Value(self, valueBytes):
		return unpack(DEFAULT_BYTES_ORDER_2+'f', valueBytes)[0]
	
	def __encodeFloat32Value(self, value):
		try:
			return pack(DEFAULT_BYTES_ORDER_2+'f', value)
		except:
			raise OverflowError
		
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_FLOAT32_BUFF_LEN, "For Float32 AVP, buffer should have 4 bytes, now it is " + str(len(valueBytes))
		return self.__decodeFloat32Value(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			if float('-inf') == avpValue or float('inf') == avpValue: raise OverflowError
			return self.__encodeFloat32Value(avpValue)
		except OverflowError as ofle:
			raise OverflowError("Float32 AVP should be b/w -3.4028234663852886E+38 - 3.4028234663852886E+38, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is float): raise ValueError("FLOAT32 AVP should have a float value. e.g. 1.7")
		return True
		
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0.0)
		
	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_FLOAT32_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPFloat32.')
		
# end of DiaAVPFloat32

class DiaAVPFloat64(DiaAVP):
	def __decodeFloat64Value(self, valueBytes):
		return unpack(DEFAULT_BYTES_ORDER_2+'d', valueBytes)[0]
	
	def __encodeFloat64Value(self, value):
		try:
			return pack(DEFAULT_BYTES_ORDER_2+'d', value)
		except:
			raise OverflowError
		
	# this function implement _decodeAVPValue of DiaAVP.																	#
	def _decodeAVPValue(self, valueBytes):
		assert type(valueBytes) is bytes
		assert len(valueBytes) is AVP_VALUE_FLOAT64_BUFF_LEN, "For Float64 AVP, buffer should have 8 bytes, now it is " + str(len(valueBytes))
		return self.__decodeFloat64Value(valueBytes)
		
	# this function implement _encodeAVPValue of DiaAVP.																	#
	def _encodeAVPValue(self, avpValue):
		try:
			if float('-inf') == avpValue or float('inf') == avpValue: raise OverflowError
			return self.__encodeFloat64Value(avpValue)
		except OverflowError as ofle:
			raise OverflowError("Float64 AVP should be b/w -1.7976931348623157E+308 - 1.7976931348623157E+308, avpValue now is "+str(avpValue))
			
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is float): raise ValueError("FLOAT64 AVP should have a float value. e.g. 1.7")
		return True

	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue(0.0)
		
	# Override __len__ operator
	def __len__(self):
		return self.getAVPHeaderLength() + AVP_VALUE_FLOAT64_BUFF_LEN
		
	#Override __getitem__ function.
	def __getitem__(self, index):
		raise NotImplementedError('__getitem__ is not implemented in DiaAVPFloat64.')
		
# end of DiaAVPFloat64
	
class DiaAVPGroup(DiaAVP):			
	# This function is used to divide a Bytes buffer which contains several AVP bytes into a list.						#
	# Each item in the list contains only 1 AVP (group AVP will contains all its sub-AVPs)								#
	@classmethod
	def cutAVPBytes(cls, avpBytes):
		LENGTH_POS = 5
		p = 0
		r = []
		while p <= len(avpBytes):
			try:
				lengthBytes = avpBytes [p + LENGTH_POS:p + LENGTH_POS + AVP_LENGTH_BUFF_LEN]
				if ZERO_LEN_BYTES == lengthBytes: break
				length = cls.decodeUIntValue(lengthBytes)
				alignedLength = cls.getAlignmentLength(length)
				r.append(avpBytes[p:p+alignedLength])
				p += alignedLength
			except Exception as e:
				break
		return r
		
	# For adding sub-AVP in this AVP object.
	def addAVP(self, avp, index = -1):
		avps = self.getAVPValue()
		assert type(avps) is list, "GROUP AVP value should be a list which contains DiaAVP sub-class objects. e.g. [avp1, avp2, avp3]"
		if -1 == index: avps.append(avp)
		elif index < len(avps): avps.insert(index, avp)
		else: raise ValueError('Index to insert is too big to this DiaAVPGroup. DiaAVPGroup only have '+ str(len(avps)) + ' items.')
		self.setAVPValue(avps)
		
	# For removing sub-AVP in this AVP object.
	def removeAVP(self, index):
		avps = self.getAVPValue()
		assert type(avps) is list
		assert index < len(avps), "Index try to remove must less than length of avps in this Group AVP object."
		del avps[index]
		self.setAVPValue(avps)

	# For getting count of sub-AVPs.
	def getAVPCount(self):
		avps=self.getAVPValue()
		assert type(avps) is list
		return len(avps)
		
	def __getAVPByCode(self, vendor, code, sameAVPIndex):
		avps = self.getAVPValue()
		i = 0
		r = None
		for avp in avps:
			if vendor == avp.getAVPVendor() and code == avp.getAVPCode(): i += 1
			if i > sameAVPIndex:
				r = avp
				break
		return r
		
	# By using DiaAVPPath object, this function can help user to visit AVP directly.									#
	# avpPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]				#
	def getAVPByPath(self, avpPath):
		assert ((type(avpPath) is DiaAVPPath),'getAVPByPath in DiaAVPGroup need a DiaAVPPath object as a parameter.')
		if 0 >= avpPath.getLayersCount(): return self
		avp = None
		vendor =  avpPath.getAVPVendor(0)
		code = avpPath.getAVPCode(0)
		sameAVPIndex = avpPath.getAVPSameAVPIndex(0)
		avp = self.__getAVPByCode(vendor, code, sameAVPIndex)
		if None == avp: return None
		if 1 < avpPath.getLayersCount():
			tmpPath = DiaAVPPath()
			tmpPath.setPath(avpPath.getPath())
			tmpPath.removeTop()
			avp = avp.getAVPByPath(tmpPath)
		return avp
		
	# By using DiaAVPPath object, this function can help user to add AVP directly.										#
	# rootAVPPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]			#
	# Note that, rootAVPPath is the father avp which contains the avp you want to add.									#
	# rootAVPPath must be a group type AVP, or you can leave rootAVPPath a blank, avp will be added at root.			#		
	def addAVPByPath(self, rootAVPPath, avp, index = -1):
		assert ((type(rootAVPPath) is DiaAVPPath),'addAVPByPath in DiaAVPGroup need a DiaAVPPath object as a parameter.')
		rootAVP = self.getAVPByPath(rootAVPPath)
		if not (type(rootAVP) is DiaAVPGroup): return -1
		rootAVP.addAVP(avp, index)
		return 0

	# By using DiaAVPPath object, this function can help user to remove AVP directly.									#
	# avpPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]				#		
	def removeAVPByPath(self, avpPath):
		assert ((type(avpPath) is DiaAVPPath),'addAVPByPath in DiaAVPGroup need a DiaAVPPath object as a parameter.')
		if 0 >= avpPath.getLayersCount(): return -1
		last = avpPath.getLayersCount() - 1
		vendor =  avpPath.getAVPVendor(last)
		code = avpPath.getAVPCode(last)
		sameAVPIndex = avpPath.getAVPSameAVPIndex(last)
		rootAVPPath = DiaAVPPath()
		rootAVPPath.setPath(avpPath.getPath())
		rootAVPPath.removeTail()
		rootAVP = self.getAVPByPath(rootAVPPath)
		if None == rootAVP: return -1
		avp = rootAVP.__getAVPByCode(vendor, code, sameAVPIndex)
		avps = rootAVP.getAVPValue()
		avps.remove(avp)
	
	# For decoding Group AVP.
	def _decodeAVPValue(self, valueBytes):
		retValue = []
		factory = DiaAVPFactory()
		avpsBytes = self.cutAVPBytes(valueBytes)
		for avpBytes in avpsBytes:
			avpObj = factory.generateDiaAVPObject(avpBytes)
			retValue.append(avpObj)
		return retValue
			
	# For encoding Group AVP.
	def _encodeAVPValue(self, avpValue):
		assert type(avpValue) is list
		retValue = ZERO_LEN_BYTES
		for subAvp in avpValue:
			retValue += subAvp.encode()
		return retValue
		
	# Override the _validAVPValue function.																				#
	def _validAVPValue(self, value):
		if (not type(value) is list): raise ValueError("GROUP AVP value should be a list which contains DiaAVP sub-class objects. e.g. [avp1, avp2, avp3]")
		for obj in value:
			if not isinstance(obj, DiaAVP): raise ValueError("GROUP AVP avlue should be a list which contains DiaAVP sub-class objects. e.g. [avp1, avp2, avp3]")
		return True
		
	# Override _initAVPValue function.
	def _initAVPValue(self):
		self.setAVPValue([])
		
	# Override __len__ operator
	def __len__(self):
		l = 0
		for subAvp in self.getAVPValue():
			l += self.getAlignmentLength(len(subAvp))
		return self.getAVPHeaderLength() + l

	#Override __getitem__ function.
	def __getitem__(self, index):
		avps = self.getAVPValue()
		assert index < len(avps), "Index try to visit must less than length of avps in this Group AVP object."
		return avps[index]

# end of DiaAVPGroup.



# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

# In tools module, all functions name use python style.																	#

from datetime import datetime, timedelta
from pyDiameter.pyDiaConst import DEFAULT_BYTES_ORDER
from pyDiameter.pyDiaConst import ZERO_LEN_BYTES
from pyDiameter.pyDiaConst import NULL_BYTE

# For address to bytes convertion.																						#
__IPV4_TYPE_STR = 'ipv4'
__IPV6_TYPE_STR = 'ipv6'
__IPV4_TYPE_VALUE = 1
__IPV6_TYPE_VALUE = 2
__IPV4_VALUE_BUFF_LEN = 4
__IPV6_VALUE_BUFF_LEN = 16
__PROTOCOL_TYPE_BUFF_LEN = 2

def __bytes_to_ipv4(bytes):
	assert len(bytes) == __IPV4_VALUE_BUFF_LEN
	r = ''
	for b in bytes:
		r += str(b)
		r += '.'
	return r[:-1]
	
def __bytes_to_ipv6(bytes, compress = 0):
	assert len(bytes) == __IPV6_VALUE_BUFF_LEN
	i = 0
	r = ''
	for b in bytes:
		r += (('%02x') %b)
		if (i+1) % 2 == 0: r += ':'
		i += 1
	r = r[:-1]
	if compress == 0: return r
	i = 0
	last_start = -1
	max_zero_len = 0
	start = 0
	end = 0
	while i < len(r):
		start = r.find('0000:', end)
		if -1 == start:
			break
		i = start
		j = i
		while j < len(r):
			if r[j] == '0' or r[j] == ':':
				j += 1
			else:
				break;
		if (j-i) > max_zero_len: 
			max_zero_len = j - i
			last_start = start
		end = j
	if 0 == last_start and max_zero_len >= 8 : r = r[:last_start] + '::' + r[last_start+max_zero_len:]
	elif 0 < last_start and max_zero_len >= 8 : r = r[:last_start] + ':' + r[last_start+max_zero_len:]
	r = r.replace('0000:', '0:')
	return r
	
def __ipv4_to_bytes(address):
	r = ZERO_LEN_BYTES
	numstrs = address.split(".")
	for numstr in numstrs:
		i = int(numstr)
		r += i.to_bytes(1, DEFAULT_BYTES_ORDER)
	return r
	
def __ipv6_to_bytes(address):
	def __hexstr_to_bytes(hexStr):
		if len(hexStr) == 1: hexStr = '0' + hexStr
		assert len(hexStr) % 2 == 0
		i = 0
		r =  ZERO_LEN_BYTES
		while i < len(hexStr):
			s = hexStr[i:i+2]
			r += int(s, 16).to_bytes(1, DEFAULT_BYTES_ORDER)
			i += 2
		return r
	i = 0
	r = ZERO_LEN_BYTES
	multiZeroIndex = address.find('::')
	if -1 == multiZeroIndex:
		address = address.replace(':','')
		r = __hexstr_to_bytes(address)
	else: 
		r1 = __hexstr_to_bytes(address[:multiZeroIndex].replace(':',''))
		r2 = __hexstr_to_bytes(address[multiZeroIndex+2:].replace(':',''))
		r = r1 + NULL_BYTE * int((__IPV6_VALUE_BUFF_LEN - len(r1) - len(r2))) + r2
	return r
	
def bytes_to_address(bytes_buff, compress = 0):
	address_type_bytes = bytes_buff[:__PROTOCOL_TYPE_BUFF_LEN]
	address_value_bytes = bytes_buff[__PROTOCOL_TYPE_BUFF_LEN:]
	if __IPV4_TYPE_VALUE == int.from_bytes(address_type_bytes,DEFAULT_BYTES_ORDER): 
		address_type = __IPV4_TYPE_STR
		address_value = __bytes_to_ipv4(address_value_bytes)
	elif __IPV6_TYPE_VALUE == int.from_bytes(address_type_bytes,DEFAULT_BYTES_ORDER): 
		address_type = __IPV6_TYPE_STR
		address_value = __bytes_to_ipv6(address_value_bytes, compress)
	return (address_type, address_value)
	
def address_to_bytes(address_tuple):
	address_type, address_value = address_tuple
	r = ZERO_LEN_BYTES
	if address_type == __IPV4_TYPE_STR: 
		r = __IPV4_TYPE_VALUE.to_bytes(2, DEFAULT_BYTES_ORDER)
		r += __ipv4_to_bytes(address_value)
	elif address_type == __IPV6_TYPE_STR: 
		r = __IPV6_TYPE_VALUE.to_bytes(2, DEFAULT_BYTES_ORDER)
		r += __ipv6_to_bytes(address_value)
	else:
		raise ValueError('Only ipv4 and ipv6 are supported by this function.')
	return r
	
# For time and bytes convertion.																									#
# Even RFC5905 obsoletes RFC4330,																									#
# but Diameter Base Protocol RFC6773 still tell us Time format should support time from year 1968 to 2104							#
# Actually, Time format could be translated to anytime.																				#
# In this tool, we will format bytes to time string according to RFC6773 Section 6. 												#
# Year 1968 Jan 20, 03:14:08 to Year 2104, Feb 26, 09:42:23																			#
# Year 1968 Jan 20, 03:14:08 is the time 2147483648(0x80000000) seconds elapse from 1900, Jan 1, 00:00:00							#
__Y1968_JAN_SECS = 2147483648
__TOTAL_SECS = 4294967296
__TIMESTAMP_BYTES_LEN = 4

def bytes_to_time(bytesBuff):
	epoc = datetime(1968, 1, 20, 3, 14, 8)
	s = int.from_bytes(bytesBuff, DEFAULT_BYTES_ORDER)
	s -= __Y1968_JAN_SECS
	if s < 0: s += __TOTAL_SECS
	dt = timedelta(seconds = s)
	r = epoc + dt
	return (r.year, r.month, r.day, r.hour, r.minute, r.second)
		
def time_to_bytes(year = 1900, month = 1, day = 1, hour = 0, minute = 0, second = 0):
	epoc = datetime(1900, 1, 1, 0, 0, 0)
	target_date = datetime(year, month, day, hour, minute, second)
	s = int((target_date - epoc).total_seconds())
	b = (s % __TOTAL_SECS)
	return b.to_bytes(__TIMESTAMP_BYTES_LEN, DEFAULT_BYTES_ORDER)



from pyDiameter.pyDiaAVPDict import DiaAVPDict
from pyDiameter.pyDiaAVPConst import AVP_TYPE_GROUP,AVP_TYPE_STR

a = DiaAVPDict()
b = DiaAVPDict()
# verify that DiaAVPDict is a singleton class.
print(a==b)

# verify whether DiaAVPDict has vendor 0 defination.
print(a.hasVendorDef(0))
# verify whether DiaAVPDict has vendor 44556 defination. (44556 is not defined in DiaAVPDict default dictionary.)
print(b.hasVendorDef(44556))

# try to get AVP type defination with code 13331, vendor code 0. No definatiion actually.
print(a.getAVPDefType(0,13331))

# add the definatiion.
a.addAVPDef(0,13331,'test',AVP_TYPE_GROUP)

# verify again by using another DiaAVPDict object.
print(b.hasAVPDef(0,13331),b.getAVPDefName(0,13331),b.getAVPDefType(0,13331))

# update the definatiion.
a.updateAVPDef(0,13331,'test2',AVP_TYPE_STR)
# verify again by using another DiaAVPDict object.
print(b.hasAVPDef(0,13331),b.getAVPDefName(0,13331),b.getAVPDefType(0,13331))



# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

# Decode Bytes Order																								#
BYTES_ORDER_BIGENDIAN = 'big'
BYTES_ORDER_LITTLEENDIAN = 'little'
DEFAULT_BYTES_ORDER = BYTES_ORDER_BIGENDIAN
BYTES_ORDER_2_BIGENDIAN = '>'
BYTES_ORDER_2_LITTLEDIAN = '<'
DEFAULT_BYTES_ORDER_2 =  BYTES_ORDER_2_BIGENDIAN

# Const for Bytes class.																							#
NULL_BYTE = b'\x00'
ZERO_LEN_BYTES = b'' 




avpBuff1 = b'\x00\x00\x05\x85\xc0\x00\x00\xa8\x00\x00\x28\xaf\x00\x00\x05\x86\xc0\x00\x00\x9c\x00\x00\x28\xaf\x00\x00\x05\x8b\xc0\x00\x00\x10\x00\x00\x28\xaf\x00\x00\x00\x01\x00\x00\x05\xa7\xc0\x00\x00\x1c\x00\x00\x28\xaf\x0b\xd5\x14\xf6\xe1\xeb\xf1\x0b\x59\xdc\xbb\x9e\x72\x90\x20\x98\x00\x00\x05\xa8\xc0\x00\x00\x1c\x00\x00\x28\xaf\x1a\xf7\x27\xb2\xb4\x8d\x86\x83\x48\xfe\x88\xda\x27\xf6\x57\x99\x00\x00\x05\xa9\xc0\x00\x00\x1c\x00\x00\x28\xaf\x01\x00\x10\x44\x41\x62\x71\x08\x11\x00\x33\x04\x50\x00\x20\x00\x00\x00\x05\xaa\xc0\x00\x00\x2c\x00\x00\x28\xaf\x12\x34\x56\x78\x90\x12\x34\x56\x78\x90\x12\x34\x56\x78\x90\x12\x34\x56\x78\x90\x12\x34\x56\x78\x90\x12\x34\x56\x78\x90\x12\x34\x00'

avpBuff2 = b'\x00\x00\x01\x25\x40\x00\x00\x11\x48\x53\x53\x2e\x74\x74\x63\x6e\x33\x00\x00\x00'

# This is a wrong avp, whose code is not defined in DiaAVPDict
avpBuff3 = b'\x00\x00\x34\x13\x40\x00\x00\x11\x48\x53\x53\x2e\x74\x74\x63\x6e\x33\x00\x00\x00'

def visitAVP(avp, tab=''):
	print(tab,end='')
	print('name:  ',avp.getAVPName())
	print(tab,end='')
	print('type:  ',avp.getAVPType())	
	print(tab,end='')
	print('code:  ',avp.getAVPCode())
	print(tab,end='')
	print('flags: ',avp.getAVPFlags())
	print(tab,end='')
	print('len:   ',len(avp))
	value=avp.getAVPValue()
	if avp.getAVPVSFlag():
		print(tab,end='')
		print('vendor:', avp.getAVPVendor())
	if type(value) is list:
		print(tab, end='')
		print('====>')
		for sub in value:
			visitAVP(sub, tab+'    ')
		print(tab, end='')
		print('<====')		
	else:
		print(tab,end='')
		print('value: ', value)
	print(tab,end='')
	print('-------')
	
from pyDiameter.pyDiaAVPFactory import DiaAVPFactory
from pyDiameter.pyDiaAVPDict import DiaAVPDict
from pyDiameter.pyDiaAVPConst import AVP_TYPE_STR

print("=========== AVP1 decoding ===========")
factory = DiaAVPFactory()
avp1 = factory.generateDiaAVPObject(avpBuff1)
visitAVP(avp1)

print("=========== AVP2 decoding ===========")
avp2 = factory.generateDiaAVPObject(avpBuff2)
visitAVP(avp2)

print("=========== AVP3 decoding ===========")
# Unknown avp decode
avp3 = factory.generateDiaAVPObject(avpBuff3)
print(avp3 == None)
# decode result is a None object.
# modify DiaAVPDict
print("=========== AVP3 Update Dictionary ===========")
dict = DiaAVPDict()
dict.addAVPDef(0,13331,'test-avp-defined',AVP_TYPE_STR)
avp3 = factory.generateDiaAVPObject(avpBuff3)
visitAVP(avp3)
print("=========== Add an int32 AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPInt32
from pyDiameter.pyDiaAVPPath import DiaAVPPath

newAVP = DiaAVPInt32()
newAVP.setAVPCode(6)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(-10)
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add an unsigned int32 AVP to AVP1's sub-avp ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPUInt32
from pyDiameter.pyDiaAVPPath import DiaAVPPath
newAVP = DiaAVPUInt32()
newAVP.setAVPCode(5)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(32)
avpPath = DiaAVPPath()
avpPath.setPath('10415/1414[0]')	# root's sub-avp (code 1414, vendor 10415)
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add an int64 AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPInt64
from pyDiameter.pyDiaAVPPath import DiaAVPPath
newAVP = DiaAVPInt64()
newAVP.setAVPCode(447)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(-47)
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add an unsigned int64 AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPUInt64
from pyDiameter.pyDiaAVPPath import DiaAVPPath
newAVP = DiaAVPUInt64()
newAVP.setAVPCode(465)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(40)
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add a float32 AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPFloat32
from pyDiameter.pyDiaAVPPath import DiaAVPPath
newAVP = DiaAVPFloat32()
newAVP.setAVPCode(496)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(1.7)
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add a float64 AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPFloat64
from pyDiameter.pyDiaAVPPath import DiaAVPPath
newAVP = DiaAVPFloat64()
newAVP.setAVPCode(603)
newAVP.setAVPMandatoryFlag()
# need set Vendor flag after setting vendor code for this AVP.
newAVP.setAVPVendor(193)
newAVP.setAVPVSFlag()
newAVP.setAVPValue(-1.18973149535723176E+308)
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add a time AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPStr
from pyDiameter.pyDiaAVPPath import DiaAVPPath
from pyDiameter.pyDiaAVPTools import address_to_bytes, bytes_to_address, time_to_bytes, bytes_to_time
import datetime
newAVP = DiaAVPStr()
newAVP.setAVPCode(55)
newAVP.setAVPMandatoryFlag()
d=datetime.datetime.now()
newAVP.setAVPValue(time_to_bytes(d.year,d.month,d.day,d.hour,d.minute,d.second))
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Add an address AVP to AVP1 ===========")
# as we known, avp1 is a group avp, only group avp could add sub-avp.
from pyDiameter.pyDiaAVPBasicTypes import DiaAVPStr
from pyDiameter.pyDiaAVPPath import DiaAVPPath
from pyDiameter.pyDiaAVPTools import address_to_bytes, bytes_to_address, time_to_bytes, bytes_to_time
newAVP = DiaAVPStr()
newAVP.setAVPCode(257)
newAVP.setAVPMandatoryFlag()
newAVP.setAVPValue(address_to_bytes(('ipv6', '2000:1234:ac12:abcd:abcd:abcd:abcd:abc1')))
avpPath = DiaAVPPath()
avpPath.setPath('')	# root position
avp1.addAVPByPath(avpPath, newAVP)
visitAVP(avp1)

print("=========== Convert avp1 object to bytes ===========")
print(avp1.encode())

# Try following functions and find the result by yourself.
avp1.setAVPMandatoryFlag()
avp1.clearAVPMandatoryFlag()
avp1.setAVPVSFlag()
avp1.clearAVPVSFlag()
avp1.setAVPProtectedFlag()
avp1.clearAVPProtectedFlag()



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='pyDiameter',
    version='0.18.7.15',
    description=(
        'pyDiameter is created by George Zhao, who is working for NOKIA, 2010-2012, 2014-2018.\n This lib is used for encoding and decoding Diameter protocol message(an AAA protocol replace the RADIUS).'
    ),
    long_description="""
===================
Introduction
===================
pyDiameter is a library which is used to decoding and encoding Diameter protocol message.

Diameter protocol is an AAA protocol and defined in RFC6733 (RFC3588 is obsoleted by RFC6733).

Decoding a diameter protocol message means, the library will help application decode protolcol bytes to the library object, which is more simple to handle.

Encoding a diameter protocol message means, the library will help application encode the library object(usually a DiaMessage object in pyDiameter library) into a bytes variable.

pyDiameter is written by python purely.

pyDiameter is written on python 3.6.5, python 2 is not supported.

===================
Author
===================
George Zhao who is working for Nokia in 2010-2012, 2014-2018.

maito: georgezhao_1980@163.com

===================
How to get it
===================

::

	pip install pyDiameter

===================
Getting start
===================
To get a start for decoding or encoding a diameter message with pyDiameter, you can read the example after pip installing.

The example folder is located in pyDiameter folder in your site-packages of your Python.

Please refer to example-dictTest.py to get start for the Diameter protocol AVP dictionary.

Dictionary
===============
This dictionary contains more than 2000 AVP definitions, as well as including several popular vendors definitions (Base, 3GPP, Nokia, Ericsson, Cisco etc.)

Please note that, decoding is depending on the dictionary defined in pyDiameter.pyDiaAVPDict, if you have some special AVP definition which is not contained in this dictionary, example-dictTest.py could help you to know how to add new AVP definition in the dictionary, and how to update it.

Please make sure that there are all definitions you need defined in the dictionary before decoding the diameter message.

AVP and Message
===============
Please refer to example-avp.py to get start for the Diameter protolcol AVP handling.

And refer to example-msg1.py to get start for the Diameter protocol message handling.

If you want an example just like a 'Hello, world', please refer to code below.

::

	from pyDiameter.pyDiaMessage import DiaMessage
	from pyDiameter.pyDiaAVPTypes import DiaAVPStr
	from pyDiameter.pyDiaAVPPath import DiaAVPPath

	msg = DiaMessage()
	msg.setRequestFlag()
	avp = DiaAVPStr()
	avp.setAVPCode(263)
	avp.setAVPVSFlag()
	avp.setAVPVendor(10415)
	avp.setAVPValue(b'Hello, world.')
	dap = DiaAVPPath()
	dap.setPath('')
	msg.addAVPByPath(dap, avp)

pyDiameter supports only basic types defined in RFC6733.

Please refer to following table for supported types list. (import them from pyDiameter.pyDiaAVPTypes)

=============== ===============
pyDiameter 		RFC6733
=============== ===============
DiaAVPStr 		OctectString
DiaAVPInt32 	Integer32
DiaAVPUInt32 	Unsigned32
DiaAVPInt64 	Integer64
DiaAVPUInt64 	Unsigned64
DiaAVPFloat32 	Float32
DiaAVPFloat64 	Float64
DiaAVPGroup 	Grouped
===============	===============

All DiaAVPxxx classes are derived from DiaAVP which is defined in pyDiameter.pyDiaAVPBase. And users could not initialize a DiaAVP object since it is an abstract class.

All Sub-classes of DiaAVP have method called getAVPType().

It return a Python string which describes the type of the object.

The returned value could be one of values of following. (import them from pyDiameter.pyDiaAVPConst)

================== ===============
Const 				Value
================== ===============
AVP_TYPE_UNKNOWN 	unknown
AVP_TYPE_STR 		str
AVP_TYPE_INT32 		int32
AVP_TYPE_UINT32 	uint32
AVP_TYPE_INT64 		int64
AVP_TYPE_UINT64 	uint64
AVP_TYPE_FLOAT32 	float32
AVP_TYPE_FLOAT64 	float64
AVP_TYPE_GROUP		grp
================== ===============

AVP Path
===============
When you operate DiaMessage object in pyDiameter.pyDiaMessage, sometimes, you need visit, add or remove some AVPs from the message.

You can use DiaAVPPath object to locate the AVP in the message.

DiaAVPPath could be imported from pyDiameter.pyDiaAVPPath.

And DiaAVPPath object uses a diameter avp path language to describe the AVP path. (Defined by George Zhao, not a well-known language.)

The AVP path is described as:

::

'vendor/avpCode[sameCodeIndex]->vendor/avpCode[sameCodeIndex]->vendor/avpCode[sameCodeIndex]...'

e.g. 10415/628[1]->10415/630[0] means I want to visit the AVP, the location is: the first AVP 630(vendor 10415) under the second AVP 628(vendor 10415).

sameCodeIndex means if there are several same AVPs with same AVP code and vendor, which index is target to be visited.

**Note that, sameCodeIndex is counted from zero.**

Tools
===============

In pyDiameter.pyDiaAVPTools, we provide four functions to help you treat some special AVP types.

The address type, and the time type are both derived from OctectString basic type according to RFC6733.

================== ================================================================================================================================
Functions			Examples
================== ================================================================================================================================
address_to_bytes	address_to_bytes(('ipv4','172.18.0.1')); address_to_bytes(('ipv6','2000:1234:ac12:abcd:abcd:abcd:abcd:abc1'))
bytes_to_address	bytes_to_address(diameter_address_avp_bytes_buff)
time_to_bytes		time_to_bytes(year=1980, month=7 , day=8, hour=20, minute=0, second=0)
bytes_to_time		bytes_to_time(bytes_buff)
================== ================================================================================================================================

**Note that, time is counted from Year 1968 Jan 20, 03:14:08 to Year 2104, Feb 26, 09:42:23, according to RFC6733 and an even obsoleted RFC4330.**

**Only ipv4 and ipv6 are supported by tools, other protocols should be handled by user as an ordinary OctectString.**

**When you set ipv6 value, the compress ipv6 value could be supported by tools, like 2000::1 or ::1.**

Any other problem, please refer to 'Any problem' section below.

===================
Any problem
===================
Please contact georgezhao_1980@163.com

Author will check mail irregularly, support is not guaranteed.

Author will answer question as immediately as author can.
""",
    author='George Zhao',
    author_email='georgezhao_1980@163.com',
    maintainer='George Zhao',
    maintainer_email='georgezhao_1980@163.com',
    license='BSD License',
    packages=find_packages(),
    platforms=["all"],
    url='http://not-available.now',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ],
)